from .error import LookupException, OutOfOettingerException
from .program import Library, Program

import sys

DEFUN = 10
FUN = DEFUN * 10

class Context:
    def __init__(self):
        self._stdin, self._stdout = sys.stdin, sys.stdout
        self._globals = {}
        self._includes = []

        self._fun = 100
        self._loaded, self._locals, self._loops = {}, [], []
        self._exit_code = 0

        self._funny_mode = True

    def __getitem__(self, key):
        self.defun()
        key = str(key)

        if self._locals and key in self._locals[-1][0]:
            return self._locals[-1][0][key]

        try:
            return self.lookup(key)

        except LookupException:
            pass

        return self._globals[key]

    def __setitem__(self, key, value):
        self.defun()
        # if there is a last stack frame, assign to it
        scope = self._locals[-1][0] if self._locals else self._globals
        scope[str(key)] = value

    def _search_file(self, name):
        import os

        name = str(name)

        for path in self._includes:
            if os.path.exists(path) and os.path.isdir(path):
                for fname in os.listdir(path):
                    front, fext = os.path.splitext(fname)
                    if not (fext == Program.FEXT or fext == Program.FEXTC):
                        continue
                    if front == name:
                        return os.path.join(path, fname)

        return None

    def stdin(self):
        return self._stdin

    def set_stdin(self, stdin):
        self._stdin = stdin

    def stdout(self):
        return self._stdout

    def set_stdout(self, stdout):
        self._stdout = stdout

    def exit_code(self):
        return self._exit_code

    def set_exit_code(self, code):
        self._exit_code = code

    def is_funny_mode(self):
        return self._funny_mode

    def disable_funny_mode(self):
        self._funny_mode = False

    def add_include_path(self, path):
        self._includes.append(path)

    def load(self, program: Program):
        for use in program.uses():
            path = self._search_file(use)
            if not path:
                raise LookupException('cannot use `{}`, not found'.format(str(use)))

            # TODO: avoid reimporting programs
            self.load(Library.load(path))

        self._loaded[program.modname()] = program

    def loaded(self):
        return self._loaded

    def lookup(self, name):
        for _, loaded in self.loaded().items():
            if name in loaded:
                return loaded[name]
        raise LookupException('cannot use `{}`, not found.'.format(name))

    def fun(self):
        if not self.is_funny_mode():
            return

        if self._fun <= 0:
            raise OutOfOettingerException
        if self._fun <= 75:
            self._fun += FUN

    def defun(self):
        if not self.is_funny_mode():
            return

        self._fun -= DEFUN
        if self._fun <= 0:
            raise OutOfOettingerException

    def push_loop(self, loop):
        self._loops.append(loop)

    def pop_loop(self):
        self._loops.pop()

    def last_loop(self):
        return self._loops[-1]

    def set_return(self, value):
        self._locals[-1][1](value)

    def push_locals(self, frame, rec_return):
        self._locals.append((frame, rec_return))

    def pop_locals(self):
        self._locals.pop()



from ..ast import *

SYN_PHRASES = [
    ('action please', MainMarker),  # main method
    ('passt so', Operator),         # rounding nums: <value> passt so 
    ('das ist', Compare),           # comparison
    ('im quadrat', Operator),       # square something
    ('und zwar', Declaration),      # function declaration: und zwar <ident> mit <ident1>
    ('und fertig', Return),         # return from a function: <value> und fertig
    ('das holen wir nach', Repeat), # repeat last statement
    ('ach kris.', Else),            # add an else branch to an if
]

SYN_OPERATORS = [
    ('+', Add),
    ('-', Sub),
    ('*', Mul),
    ('/', Div),
    ('modulo', Mod),
    ('<', Less),
]

SYN_KEYWORDS = [
    ('use', Use),               # import functions from a file
    ('uzbl', UzblConstant),     # True
    ('menge', Menge),           # object-like thing
    ('liste', Liste),           # list-like thing
    ('quasi', Print),           # output value
    ('bitte?', Readin),         # input value
    ('und', LogicalAnd),        # logical and; concatenate strings (?)
    ('oder', LogicalOr),        # logical or
    ('not', LogicalNot),        # logical not
    ('so', SoMarker),           # speed up program
    ('stark', ConstMarker),     # declare a constant
    ('mit', Arguments),         # introduce arguments
    ('also', RHAssign),         # right hand assignment <val> = <ident>
    ('ist', LHAssign),          # left hand assignment <ident> = <val>
    ('jens', Exit),             # exit program
    ('kris?', If),              # if
    ('kris??', Elif),           # else if
    ('patrick!', Break),        # break
    ('hä', Debug),              # acts as breakpoint for debugger
    ('softwareproblem', Raise), # raise error
    ('fähler', Raise),          # read last error
    ('oettinger', Nop),         # keine ahnung
]

KEYWORDS = []
SYN_TREE = {}

def setup():
    global KEYWORDS
    global SYN_TREE

    # add keywords
    KEYWORDS = list(map(lambda x: x[0], SYN_KEYWORDS))

    # add operators
    KEYWORDS.extend(list(map(lambda x: x[0], SYN_OPERATORS)))

    for parts, tag in map(lambda x: (x[0].split(' '), x[1]), SYN_PHRASES):
        node = SYN_TREE
        for k in parts:
            if k not in KEYWORDS:
                KEYWORDS.append(k)

            if k not in node:
                new_node = {}
                node[k] = new_node
                node = new_node
            else:
                node = node[k]
        node['_op'] = tag

    for k, tag in SYN_KEYWORDS:
        if k not in SYN_TREE:
            SYN_TREE[k] = { '_op': tag }
        else:
            SYN_TREE[k]['_op'] = tag

    for k, tag in SYN_OPERATORS:
        if k not in SYN_TREE:
            SYN_TREE[k] = { '_op': tag }
        else:
            SYN_TREE[k]['_op'] = tag

setup()



from .ast import *
from .ast.generic import *

import dill
import inspect
import re
import sys

def create_fn(fn):
    sig = inspect.signature(fn)
    param_protocols = []

    # create a list of (param_name, conv_fn)-pairs where 
    # conv_fn is called on values retrieved from the context
    for param_name, param in sig.parameters.items():
        conv_fn = None
        if param.annotation != inspect._empty:
            conv_fn = param.annotation
        param_protocols.append((param_name, conv_fn))

    return PyFunction(param_protocols, fn)

#def create_fn_with_context(name, fn):
#    return PyFunction(name, fn)

def create_const(name, value):
    return Value.create(value)

def get_vlib_modname_by_path(path):
    from os.path import abspath, basename, splitext
    fname = basename(abspath(path))
    fname, _ = splitext(fname)
    return get_vlib_modname(fname)

def get_vlib_modname(sub=None):
    if sub is None:
        return Library.VIRTUAL_MODULE
    return '{}.{}'.format(Library.VIRTUAL_MODULE, sub)

def _load_binary(fname):
    modname = get_vlib_modname_by_path(fname)
    init_vlib()
    init_vlib(modname)

    with open(fname, 'rb') as fp:
        sys.modules[modname] = dill.load(fp)

    return sys.modules.get(modname)

def _load_source(fname):
    from .generate import Compiler
    compiler = Compiler()
    with open(fname, 'r') as src:
        return compiler.compile(src.read())

def init_vlib(modname=None, mod=None):
    """
    this initializes a virtual module containing all loaded qc library files.
    """
    import types

    if modname is None:
        modname = get_vlib_modname()

    if modname in sys.modules and modname == Library.VIRTUAL_MODULE:
        # this makes 'vlib' a package
        sys.modules[modname].__path__ = iter([])
    
    else:
        if mod is None:
            mod = types.ModuleType(modname, 'virtual module for importing libraries')

        if modname == Library.VIRTUAL_MODULE:
            # this makes 'vlib' a package
            mod.__path__ = iter([])

        sys.modules[modname] = mod

class Library(object):
    FEXT = '.qc'
    FEXTC = '.qcc'
    VIRTUAL_MODULE = '__main__'

    def __init__(self):
        self._idents = {}
        self._uses = []
        self._file = None

    def __contains__(self, key):
        return key in self._idents

    def __getitem__(self, key):
        if key not in self._idents:
            return None
        return self._idents[key]

    def set_file_path(self, fname):
        from os.path import abspath
        self._file = fname

    # load a program from filepointer
    def load(fname):
        from os.path import splitext

        _, fext = splitext(fname)

        if fext == Library.FEXT:
            lib = _load_source(fname)

        elif fext == Library.FEXTC:
            lib = _load_binary(fname)

        else:
            raise Exception('unsupported file extension')

        lib.set_file_path(fname)

        return lib

    # write a program into filepointer
    def save(self, fname):
        with open(fname, 'wb') as fp:
            dill.dump(self, fp, recurse=True)
        return True

    def ident(self, ident: str, value):
        self._idents[ident] = value

    def idents(self) -> dict:
        return self._idents

    def use(self, name):
        self._uses.append(name)

    def uses(self) -> list:
        return self._uses

    def modname(self):
        if self._file is None:
            return None

        from os.path import basename, splitext
        fname, _ = splitext(basename(self._file))
        return fname

    def __str__(self) -> str:
        fixed = map(lambda x: '{}:\n{}'.format(x[0], x[1]), self._idents.items())
        return '\n'.join(fixed)

class PyLibrary(Library):
    def __init__(self):
        Library.__init__(self)

class PyFunction(Function):
    def __init__(self, args, fn):
        super().__init__(args, Block())
        self._fn = fn

    def args(self):
        return [name for name, _ in self._args]

    def block(self):
        return self

    def __str__(self):
        return '<python function>'

    def run(self, ctx):
        kwargs = {}

        # generate a **kwargs out of ctx when called; pass it to `fn`
        for param_name, conv_fn in self._args:
            kwargs[param_name] = ctx[param_name]
            if not conv_fn is None:
                kwargs[param_name] = conv_fn(kwargs[param_name])

        return self._fn(**kwargs)



from .ast import *
from .ast.generic import isof
from .context import *
from .program import *

class Interpreter:
    LIB_PATH = '/usr/local/lib/quasicode'
    USERLIB_PATH = '~/.local/lib/quasicode'

    def __init__(self):
        import os

        self._ctx = Context()
        self._ctx.add_include_path(os.getcwd())
        self._ctx.add_include_path(os.path.expanduser(Interpreter.USERLIB_PATH))
        self._ctx.add_include_path(Interpreter.LIB_PATH)

    def disable_funny_mode(self):
        self._ctx.disable_funny_mode()

    def load(self, program: Program):
        self._ctx.load(program)

    def _run_func(self, func: Function, ctx: Context):
        last = None
        for step in func.block():
            last = step.run(ctx)
        return last

    def call(self, name: str, args: list=[]):
        args = [Value.create(arg) for arg in args]
        call = FunctionCall(Ident(name), args)
        return call.run(self._ctx)

    def run(self):
        try:
            for _, loaded in self._ctx.loaded().items():
                if not isof(loaded, Program):
                    continue
                ep = loaded.entry_point()
                if ep:
                    self._run_func(loaded[ep], self._ctx)
                    break
            else:
                raise Exception('no starting point')

        except Exception as e:
            import traceback
            self._ctx.set_exit_code(1)
            print(traceback.format_exc())
            print(e)



def isof(var, cls) -> bool:
    return isinstance(var, cls) or any(isof(cls, b) for b in var.__class__.__bases__ if b.__name__ != 'object')

def decl_key(k):
    if isinstance(k, str):
        return k
    if hasattr(k, 'name'):
        return k.name()
    raise Exception('variable `{}` is not valid as function argument'.format(k))

class Runnable:
    def run(self, ctx):
        pass

class Parameterized:
    def __init__(self):
        self._args = []
    
    def args(self):
        return self._args

    def add_arg(self, arg):
        self._args.append(arg)

class Keyword:
    def __init__(self, name=None):
        self._name = name
        if self._name is None:
            self._name = self.__class__.__name__

    def name(self):
        return self._name

    def eq(self, other):
        if isinstance(other, str):
            return self._name == other
        if isof(other, Keyword):
            return self._name == other._name
        return False

    def __str__(self):
        return self.name()

class Arguments(Keyword):
    def __init__(self):
        super().__init__()

class Statement(Keyword):
    def __init__(self):
        super().__init__()

class NestedStatement(Statement, Runnable):
    def __init__(self):
        self._block = Block()

    def block(self):
        return self._block

    def set_block(self, block):
        self._block = block

    def run(self, ctx):
        for step in self._block:
            step.run(ctx)

    def __str__(self):
        return '\n'.join(map(str, self._block))

class Block(list, Runnable):
    def __init__(self):
        self._broken = False

    def __str__(self):
        return ' '.join(map(str, self))

    def end(self):
        self._broken = True

    def run(self, ctx):
        self._broken = False
        last = None
        for step in self:
            if self._broken:
                break
            last = step.run(ctx)
        return last

class FunctionCall(Runnable, Parameterized):
    def __init__(self, name, args=[]):
        Parameterized.__init__(self)
        self._name = name
        for arg in args:
            self.add_arg(arg)

    def __str__(self):
        return '{} {}'.format(self.name(), ' '.join(map(str, self._args)))

    def name(self):
        return self._name.name()

    def run(self, ctx):
        rvars = [arg.run(ctx) for arg in self.args()]
        decl = ctx.lookup(self.name())

        if len(decl.args()) != len(rvars):
            raise Exception('call to `{}` expected {} arguments, got {}'.format(self.name(), len(decl.args()), len(rvars)))

        frame = {decl_key(k): v for k, v in zip(decl.args(), rvars)}
        self._ret = None

        def receive_return(v):
            self._ret = v
            decl.block().end()

        ctx.push_locals(frame, receive_return)
        last = decl.block().run(ctx)
        ctx.pop_locals()

        if None != self._ret:
            return self._ret
        return last



from .ast.generic import Block, isof

class Function(object):
    def __init__(self, args, block):
        assert isof(args, list)
        assert isof(block, Block)
        self._args = args
        self._block = block

    def args(self):
        return self._args

    def block(self):
        return self._block

    def __str__(self):
        return '\n'.join(map(str, self._block))



from .library import *

class Program(Library):
    def __init__(self):
        super().__init__()
        # TODO: is flow needed?
        self._flow = []
        self._entry_point = None

    def __iter__(self):
        return iter(self._flow)

    def entry_point(self):
        return self._entry_point

    def set_entry_point(self, ident: str):
        self._entry_point = ident



import pytest

from .deps import *

@pytest.fixture(scope='session')
def srcdir(tmpdir_factory):
    d = tmpdir_factory.mktemp('tmp')

    fname = d.join('test.qc')
    src = """
und zwar gib_aus
    quasi "hallo welt"
    """
    fname.write(src)

    return d

class TestProgram(Test):
    def test_saving(self, srcdir, internals):
        fname = srcdir.join('test.qc')
        with open(str(fname), 'r') as fp:
            src = fp.read()

        program = internals.compiler.compile(src)
        self.assertTrue(program.save('{}c'.format(fname)))

    def test_loading(self, srcdir):
        fname = srcdir.join('test.qcc')
        loaded_prog = Program.load(str(fname))
        self.assertIsInstance(loaded_prog, Program)
        self.assertTrue(loaded_prog._file.endswith('test.qcc'))
        self.assertEqual('test', loaded_prog.modname())



import unittest

from .deps import *

class TestObjects(Test):
    def test_creation(self, internals):
        src = """
und zwar create_indirect
    obj1 ist liste
    obj1 und fertig

und zwar create
    liste und fertig
        """
        program = internals.compiler.compile(src)
        internals.interpreter.load(program)

        ret = internals.interpreter.call('create')
        self.assertIsInstance(ret, Liste)

        ret = internals.interpreter.call('create_indirect')
        self.assertIsInstance(ret, Liste)

    def test_creation_parameterized(self, internals):
        src = """
und zwar create_indirect
    a ist 1
    b ist 3
    obj1 ist liste mit a 2 b 4
    obj1 und fertig
        """
        program = internals.compiler.compile(src)
        internals.interpreter.load(program)

        ret = internals.interpreter.call('create_indirect')
        self.assertIsInstance(ret, Liste)
        self.assertTrue(ret._val)

        for i in range(4):
            self.assertEqual(i+1, float(ret._val[i]))

    def test_creation_inline(self, internals):
        src = """
und zwar create_indirect
    (liste mit 1 2 3) und fertig
        """
        program = internals.compiler.compile(src)
        internals.interpreter.load(program)

        ret = internals.interpreter.call('create_indirect')
        self.assertIsInstance(ret, Liste)

        for i in range(3):
            self.assertEqual(i+1, float(ret._val[i]))



from more_itertools import peekable
import re

from ..ast import *

from .error import ParserError
from .syntax import *

peek_is = lambda cls: lambda it: isof(it.peek(), cls)

def take_while(it, pred: callable):
    collect = []
    while it and pred(it):
        collect.append(next(it))
    return collect

class Indent:
    def __init__(self, depth=0):
        self._depth = int(depth)

    def depth(self):
        return self._depth

    def decrease_depth(self):
        self._depth -= 1
    
    def __str__(self):
        return '\\t{}'.format(self._depth)

class Newline:
    def __str__(self):
        return '\\n'

class Parens:
    def __init__(self, expr):
        self._expr = expr

    def expr(self):
        return self._expr

class Lexer:
    INDENT_RE = re.compile('^(\\t|\s{4})+')
    INDENT_WIDTH = 4

    def _keywords_to_ast(self, stc):
        it = iter(stc)
        node = SYN_TREE[next(it)]
        name = None

        for k in it:
            try:
                node = node[k]
            except KeyError:
                possible = ','.join(map(lambda x: '`{}`'.format(x), node.keys()))
                raise ParserError('expected one of {}. got `{}`'.format(possible, k))

        if '_op' not in node:
            raise ParserError('unexpected symbol `{}` in `{}`'.format(name, stc))

        return node['_op']

    def _interpret_buffer(self, buf: str):
        if buf == 'uzbl':
            return UzblConstant()
        elif buf == 'menge':
            return Menge()
        elif buf == 'liste':
            return Liste()
        elif buf in KEYWORDS:
            return Keyword(buf)
        elif buf.isnumeric():
            return Number(float(buf))
        else:
            parts = buf.split('.')
            if 1 < len(parts):
                return Access(parts)
            else:
                return Ident(buf)

    def _reduce_lex(self, it: peekable):
        reduced = []
        for lexem in it:
            if isof(lexem, Keyword):
                stc = take_while(it, lambda it: isof(it.peek(), Keyword))
                stc = [lexem.name(), *map(lambda kw: kw.name(), stc)]
                kw = self._keywords_to_ast(stc)
                reduced.append(kw())

            else:
                reduced.append(lexem)
        
        return reduced

    def _lex_source(self, it: peekable):
        lexed = []


        buf = ''
        for c in it:
            if c == ' ':
                pass

            elif c == '(':

                def parens_parser(init):
                    paren_stack = init
                    def predicate(it):
                        p = it.peek()
                        if p == '(':
                            paren_stack.append(p)
                        elif p == ')':
                            paren_stack.pop()
                            return 0 < len(paren_stack)
                        return True
                    return predicate

                sub = take_while(it, parens_parser(['(']))
                if it:
                    # drop trailing )
                    assert next(it) == ')'
                sub = self._lex_source(peekable(sub))
                lexed.append(Parens(sub))
                continue

            # parse strings here
            # TODO: discourage use of ' here
            elif c == '"' or c == "'":
                buf = take_while(it, lambda it: it.peek() != c)
                buf = ''.join(buf)
                lexed.append(String(buf))
                # drop trailing ' or "
                next(it)
                buf = ''
                continue

            else:
                buf += c
                continue

            if buf:
                out = self._interpret_buffer(buf)
                lexed.append(out)
                buf = ''

        if buf:
            out = self._interpret_buffer(buf)
            lexed.append(out)

        return self._reduce_lex(peekable(lexed))

    def lex(self, line: str):
        lexed = []

        match = self.INDENT_RE.match(line)
        if match:
            depth = match.span()[1]
            lexed.append(Indent(depth / self.INDENT_WIDTH))

        it = peekable(line.strip())
        lexed.extend(self._lex_source(it))

        return lexed

class Parser:
    def __init__(self):
        self._lexer = Lexer()
    
    def _parse_keyword(self, it):
        if isinstance(it, list):
            it = peekable(it)
        stc = take_while(it, peek_is(Keyword))
        if not stc:
            return None
        assert len(stc) == 1
        return stc.pop()

    def _take_block(self, lines, min_depth = 1):
        block = []
        while lines:
            next_line = lines.peek()

            # skip line if empty
            if not next_line:
                next(lines)
                continue

            if isof(next_line, Block):
                return next_line

            start = next_line[0]

            if not isof(start, Indent) or not min_depth <= start.depth():
                break

            next_line = next(lines)

            # drop indent
            next_line[0].decrease_depth()
            if next_line[0].depth() == 0:
                block.append(next_line[1:])
            else:
                block.append(list(next_line))
        return block

    def _parse_expression(self, line):
        stack = []
        if not isinstance(line, peekable):
            line = peekable(line)

        while line:
            token = next(line)

            if isof(token, Operator):
                if stack:
                    top = stack.pop()
                    if not (isof(top, Value) or isof(top, Operator)):
                        assert False
                    token.add_arg(top)
                
                stack.append(token)

            elif isof(token, Value):
                if stack:
                    if isof(stack[-1], Ident):
                        ident = stack.pop()
                        call = FunctionCall(ident)
                        call.add_arg(token)
                        stack.append(call)
                    else:
                        assert len(stack[-1].args()) < 2
                        stack[-1].add_arg(token)
                else:
                    stack.append(token)

            elif isof(token, Parens):
                sub = self._parse_expression(token.expr()) 
                if stack:
                    assert len(stack[-1].args()) < 2
                    stack[-1].add_arg(sub)
                else:
                    stack.append(sub)

            elif isof(token, Arguments):
                top = stack.pop()
                assert isof(top, Liste)
                pred = lambda it: peek_is(Value)(it) or peek_is(Ident)(it)
                args = take_while(line, pred)
                stack.append(Construct(top, args))

            elif isof(token, Readin):
                stack.append(token)

            else:
                raise ParserError('token `{}` not expected in this position'.format(token))

        if len(stack) != 1:
            print(stack)
            assert False

        return stack.pop()

    def _parse_declaration(self, line):
        name = next(line)
        args, markers = [], []

        while line:
            for kw in take_while(line, peek_is(Keyword)):
                if isof(kw, Arguments):
                    args = take_while(line, peek_is(Ident))
                elif isof(kw, Marker):
                    markers.append(kw)

        return (name, args, markers)

    def _parse_line(self, line, lines):
        line = peekable(line)
        item = next(line)

        if isof(item, Declaration):
            name, args, markers = self._parse_declaration(line)

            block = self._take_block(lines)
            block = self._parse_lines(block)

            item.set_name(name)
            item.set_args(args)
            item.set_block(block)

            for marker in markers:
                item.add_marker(marker)

        elif isof(item, Repeat):
            block = self._take_block(lines)
            block = self._parse_lines(block)
            item.set_block(block)

        elif isof(item, Branch):
            assert isof(item, If)

            condition = self._parse_expression(line)
            block = self._take_block(lines)
            block = self._parse_lines(block)

            item.add_branch(condition, block)

            if lines:
                cls = self._parse_keyword(lines.peek())
                try:
                    while isof(cls, Elif):
                        line = next(lines)
                        # drop elif keyword
                        line = line[1:]
                        condition = self._parse_expression(line)
                        block = self._take_block(lines)
                        block = self._parse_lines(block)
                        item.add_branch(condition, block)

                        cls = self._parse_keyword(lines.peek())
                    
                    if isof(cls, Else):
                        line = next(lines)
                        block = self._take_block(lines)
                        block = self._parse_lines(block)
                        item.set_default_branch(block)

                except StopIteration:
                    pass

        elif isof(item, Print) or isof(item, Readin):
            for arg in line:
                item.add_arg(self._parse_expression([arg]))

        elif isof(item, Use):
            for arg in line:
                assert isof(arg, Ident)
                item.add_arg(arg)

        elif isof(item, Value) or isof(item, Parens):
            if isof(item, Ident) and line and isof(line.peek(), Value):
                args = take_while(line, peek_is(Value))
                return FunctionCall(item, args)

            if isof(item, Parens):
                item = self._parse_expression(item.expr())

            kw = self._parse_keyword(line)
            if kw != None:
                if isof(kw, Return):
                    kw.add_arg(item)

                elif isof(kw, Statement):
                    val = self._parse_expression(line)

                    if isof(kw, LHAssign) and isof(item, Ident):
                        kw.set_ident(item)
                        kw.set_value(val)
                    elif isof(kw, RHAssign) and isof(val, Ident):
                        kw.set_ident(val)
                        kw.set_value(item)
                    else:
                        assert False

                return kw

            else:
                return item

        # drop empty indents
        elif isof(item, Indent):
            return None

        return item

    def _parse_lines(self, lines):
        block = Block()
        lines = peekable(lines)

        for line in lines:
            # skip empty lines
            if not line:
                continue

            if not isinstance(line, peekable):
                line = peekable(line)

            parsed = self._parse_line(line, lines)
            if not parsed is None:
                block.append(parsed)

        return block

    def parse(self, content: str) -> list:
        lexed = [peekable(self._lexer.lex(line)) for line in content.split('\n')]
        return self._parse_lines(lexed)



import pytest

from .deps import *

class TestParsing(Test):
    def test_declare(self, internals):
        src = """
und zwar hier action please
    quasi 1
        """
        program = internals.parser.parse(src)

        self.assertIsInstance(program[0], Declaration)
        self.assertTrue(program[0].is_main())

    def test_declare_args(self, internals):
        src = """
und zwar hier mit a b
    quasi 1
        """
        program = internals.parser.parse(src)

        self.assertIsInstance(program[0], Declaration)
        self.assertEqual(2, len(program[0].args()))

    def test_declare_args_main(self, internals):
        src = """
und zwar hier mit a b action please
    quasi 1
        """
        program = internals.parser.parse(src)

        self.assertIsInstance(program[0], Declaration)
        self.assertEqual(2, len(program[0].args()))
        self.assertTrue(program[0].is_main())

    def test_branch(self, internals):
        src = """
kris? 1 das ist 1
    quasi "lol"
        """
        program = internals.parser.parse(src)
        self.assertIsInstance(program[0], If)

    def test_branch_elif(self, internals):
        src = """
kris? 3 das ist 1
    quasi "fizz"
kris?? 3 das ist 2
    quasi "buzz"
kris?? 3 das ist 3
    quasi "buzz"
        """
        program = internals.parser.parse(src)
        self.assertIsInstance(program[0], If)
        self.assertEqual(3, len(program[0].branches()))

    def test_branch_elif_else(self, internals):
        src = """
kris? 1 das ist 2
    quasi "fizz"
kris?? 1 das ist 1
    quasi "buzz"
ach kris.
    quasi "nix"
        """
        program = internals.parser.parse(src)
        self.assertIsInstance(program[0], If)
        self.assertEqual(3, len(program[0].branches()))

    def test_repeat(self, internals):
        src = """
0 also i
das holen wir nach
    i ist i + 1
    kris? i das ist 10
        patrick!
        """
        program = internals.parser.parse(src)
        self.assertIsInstance(program[0], Assign)
        self.assertIsInstance(program[1], Repeat)

    def test_string_parsing(self, internals):
        src = """
"hello world" also str1
str2 ist 'hallo welt'
        """
        program = internals.parser.parse(src)
        self.assertIsInstance(program[0], RHAssign)
        self.assertIsInstance(program[1], LHAssign)

    def test_parens(self, internals):
        src = """
i ist (0 + 1)
        """
        program = internals.parser.parse(src)
        self.assertIsInstance(program[0], LHAssign)



from .generic import *
from .value import *

def foldtwo(fn, args, ctx):
    arg1 = args[0].run(ctx)
    arg2 = args[1].run(ctx)
    return fn(float(arg1), float(arg2))

class Operator(Keyword, Parameterized):
    def __init__(self):
        super().__init__()
        Parameterized.__init__(self)

    def _repr(self):
        return ''

    def __str__(self):
        return '({} {})'.format(self._repr(), ' '.join(map(str, self._args)))

class Compare(Operator):
    def __init__(self):
        super().__init__()

    def _repr(self):
        return '=='

    def run(self, ctx):
        return self._args[0].run(ctx) == self._args[1].run(ctx)

class Less(Operator):
    def __init__(self):
        super().__init__()

    def _repr(self):
        return '<'

    def run(self, ctx):
        return float(self._args[0].run(ctx)) < float(self._args[1].run(ctx))

class LogicalAnd(Operator):
    def __init__(self):
        super().__init__()

    def _repr(self):
        return 'und'

    def run(self, ctx):
        return self._args[0].run(ctx) and self._args[1].run(ctx)

class LogicalNot(Operator):
    def __init__(self):
        super().__init__()

    def _repr(self):
        return 'not'

    def run(self, ctx):
        return not self._args[0].run(ctx)

class LogicalOr(Operator):
    def __init__(self):
        super().__init__()

    def _repr(self):
        return 'oder'

    def run(self, ctx):
        if self._args[0].run(ctx):
            return Value.create(True)
        if self._args[1].run(ctx):
            return Value.create(True)
        return Value.create(False)

class Add(Operator):
    def __init__(self):
        super().__init__()

    def _repr(self):
        return '+'

    def run(self, ctx):
        return foldtwo(lambda a, b: a + b, self._args, ctx)

class Sub(Operator):
    def __init__(self):
        super().__init__()

    def _repr(self):
        return '-'

    def run(self, ctx):
        return foldtwo(lambda a, b: a - b, self._args, ctx)

class Mul(Operator):
    def __init__(self):
        super().__init__()

    def _repr(self):
        return '*'

    def run(self, ctx):
        return foldtwo(lambda a, b: a * b, self._args, ctx)

class Div(Operator):
    def __init__(self):
        super().__init__()

    def _repr(self):
        return '/'

    def run(self, ctx):
        return foldtwo(lambda a, b: a / b, self._args, ctx)

class Mod(Operator):
    def __init__(self):
        super().__init__()

    def _repr(self):
        return 'modulo'

    def run(self, ctx):
        return foldtwo(lambda a, b: a % b, self._args, ctx)



from .generic import *
from .value import Access

class Assign(Statement, Runnable):
    def __init__(self):
        self._ident = None
        self._value = None

    def ident(self):
        return self._ident

    def set_ident(self, ident):
        self._ident = ident

    def value(self):
        return self._value

    def set_value(self, value):
        self._value = value

    def run(self, ctx):
        if isof(self._ident, Access):
            menge = ctx[self._ident]
            last = self._ident.subparts()[-1]

            for subpart in self._ident.subparts()[:-1]:
                menge = menge[subpart]

            menge[last] = self._value.run(ctx)
        else:
            ctx[self._ident] = self._value.run(ctx)

    def __str__(self):
        return '{} = {}'.format(self._ident, self._value)

class LHAssign(Assign):
    def __init__(self):
        super().__init__()

class RHAssign(Assign):
    def __init__(self):
        super().__init__()



import pytest

from .deps import *

class TestConstants(Test):
    def test_uzbl(self, internals):
        src = """
und zwar teste
    a ist uzbl
    b ist (not uzbl)
    kris? a das ist b
        1 und fertig
    0 und fertig
        """
        program = internals.compiler.compile(src)
        internals.interpreter.load(program)

        ret = internals.interpreter.call('teste', [])
        self.assertIsInstance(ret, Number)
        self.assertEqual(float(0), float(ret))



import pytest

from .deps import *

class TestInterpreter(Test):
    def test_maths(self, internals):
        src = """
und zwar const
    42 und fertig

und zwar rechne mit x
    (x + 1) und fertig
        """
        program = internals.compiler.compile(src)
        internals.interpreter.load(program)

        self.assertEqual(42, float(internals.interpreter.call('const')))
        self.assertEqual(2, float(internals.interpreter.call('rechne', [1])))

    def test_logical(self, internals):
        src = """
und zwar or_wahr
    ((not uzbl) oder uzbl) und fertig

und zwar or_falsch
    ((not uzbl) oder (not uzbl)) und fertig

und zwar and_wahr
    (uzbl und uzbl) und fertig

und zwar and_falsch
    ((not uzbl) und uzbl) und fertig
        """
        program = internals.compiler.compile(src)
        internals.interpreter.load(program)

        self.assertTrue(internals.interpreter.call('or_wahr'))
        self.assertFalse(internals.interpreter.call('or_falsch'))

        self.assertTrue(internals.interpreter.call('and_wahr'))
        self.assertFalse(internals.interpreter.call('and_falsch'))

    def test_comparison(self, internals):
        src = """
und zwar kleiner mit a b
    kris? a < b
        uzbl und fertig
    (not uzbl) und fertig
        """
        program = internals.compiler.compile(src)
        internals.interpreter.load(program)

        self.assertTrue(internals.interpreter.call('kleiner', [1, 2]))
        self.assertFalse(internals.interpreter.call('kleiner', [2, 1]))
        self.assertFalse(internals.interpreter.call('kleiner', [2, 2]))

    def test_unknown_use(self, internals):
        src = """use non_existent"""
        program = internals.compiler.compile(src)
        with pytest.raises(LookupException):
            internals.interpreter.load(program)



import pytest

from .deps import *

class TestCompiler(Test):
    def test_unexpected_toplevel(self, internals):
        src = """
quasi "test"
        """
        with pytest.raises(CompilerError):
            program = internals.compiler.compile(src)

    def test_double_main(self, internals):
        src = """
und zwar main2 action please
    quasi "test"

und zwar main1 action please
    quasi "test"
        """
        with pytest.raises(CompilerError):
            program = internals.compiler.compile(src)



import unittest

from .deps import *

class TestObjects(Test):
    def test_creation(self, internals):
        src = """
und zwar create_indirect
    obj1 ist menge
    obj1 und fertig

und zwar create
    menge und fertig
        """
        program = internals.compiler.compile(src)
        internals.interpreter.load(program)

        ret = internals.interpreter.call('create')
        self.assertIsInstance(ret, Menge)

        ret = internals.interpreter.call('create_indirect')
        self.assertIsInstance(ret, Menge)

    def test_object_access(self, internals):
        src = """
und zwar return
    obj1 ist menge
    obj1.b ist 1

    obj1.sub ist menge
    obj1.sub.a ist 2

    obj1 und fertig
        """
        program = internals.compiler.compile(src)
        internals.interpreter.load(program)

        ret = internals.interpreter.call('return')
        self.assertEqual(1, float(ret['b']))
        self.assertEqual(2, float(ret['sub']['a']))

    def test_object_passing(self, internals):
        src = """
und zwar select_one mit m
    m.b und fertig

und zwar give
    obj1 ist menge
    obj1.a ist 1
    obj1.b ist 2
    obj1.c ist 3

    (select_one obj1) und fertig
        """
        program = internals.compiler.compile(src)
        internals.interpreter.load(program)

        ret = internals.interpreter.call('give')
        self.assertEqual(2, float(ret))

    def test_object_number_key(self, internals):
        src = """
und zwar give
    obj1 ist menge
    obj1.1 ist "first"
    obj1.2 ist "second"
    obj1.3 ist "third"

    obj1 und fertig
        """
        program = internals.compiler.compile(src)
        internals.interpreter.load(program)

        ret = internals.interpreter.call('give')
        self.assertEqual('second', str(ret['2']))



from .generic import FunctionCall, isof
from ..function import Function

class Value:
    def __init__(self, val):
        self._val = val

    def create(val):
        if isinstance(val, int) or isinstance(val, float):
            return Number(float(val))
        if isinstance(val, bool):
            if val:
                return UzblConstant()
            else:
                op = LogicalNot()
                op.add_arg(UzblConstant())
                return op
        if isinstance(val, str):
            return String(val)
        if isinstance(val, bytes):
            return String(val.decode('utf-8'))
        raise Exception('cannot create value from `{}`'.format(val))

    def __str__(self):
        return self._val

    def __int__(self):
        return self._val

    def __float__(self):
        return self._val

    def __bool__(self):
        return bool(self._val)

    def is_assignable(self):
        return False
    
    def __eq__(self, other):
        if other is None:
            return False
        if isinstance(other, float) or isinstance(other, int):
            return self._val == other
        return self._val == other._val

    def run(self, ctx):
        return self._val

class Constant(Value):
    def __init__(self, val):
        super().__init__(val)

class UzblConstant(Constant):
    def __init__(self):
        super().__init__(True)

    def __str__(self):
        return 'uzbl'

class Number(Value):
    def __init__(self, val):
        super().__init__(val)

    def run(self, ctx):
        return self

    def __str__(self):
        return str(self._val)

class Menge(Value):
    def __init__(self):
        super().__init__({})

    def __getitem__(self, key):
        return self._val[key]

    def __setitem__(self, key, value):
        self._val[key] = value

    def run(self, ctx):
        return self

    def __str__(self):
        return str(self._val)

class Liste(Value):
    def __init__(self):
        super().__init__([])

    def run(self, ctx):
        return self

    def __str__(self):
        return str(self._val)

class String(Value):
    def __init__(self, val: str):
        super().__init__(val)

class Ident(Value):
    def __init__(self, name: str):
        super().__init__(name)

    def name(self):
        return self._val

    def run(self, ctx):
        val = ctx[self._val]
        # TODO: execute FunctionCall
        if isof(val, Function):
            return FunctionCall(self, []).run(ctx)
        return val
    
    def is_assignable(self):
        return True

class Access(Ident):
    def __init__(self, parts: list):
        super().__init__(parts[0])
        self._subparts = parts[1:]

    def name(self):
        return self._val

    def subparts(self):
        return self._subparts

    def run(self, ctx):
        val = ctx[self._val]
        assert isof(val, Menge)

        for subpart in self._subparts:
            val = val[subpart]

        return val



from .generic import *
from .marker import *
from .value import *

class Declaration(NestedStatement):
    def __init__(self):
        super().__init__()
        self._is_main = False
        self._args = []
        self._name = None

    def name(self):
        return self._name.name()

    def set_name(self, name: Ident):
        self._name = name

    def args(self):
        return self._args

    def set_args(self, args):
        self._args = args

    def is_main(self):
        return self._is_main

    def add_marker(self, marker: Marker):
        if isof(marker, MainMarker):
            self._is_main = True

    def __str__(self):
        return 'Declaration'



import io
import os
import pytest

from .deps import *

def patch_path(*args):
    return os.path.join(os.getcwd(), *args)

def load_source(fname):
    with open(fname, 'r') as src:
        return src.read()

class TestExamples(Test):
    def test_all(self, internals):
        internals.interpreter._ctx.set_stdin(io.StringIO())

        examples = os.listdir(patch_path('examples'))
        ignore = ['bot.qc']

        for example in examples:
            if example in ignore:
                continue

            src_path = patch_path('examples', example)
            print('$', example)

            src = load_source(src_path)
            program = internals.compiler.compile(src)

            internals.interpreter.disable_funny_mode()

            internals.interpreter.load(program)
            internals.interpreter.run()

            self.assertEqual(0, internals.interpreter._ctx.exit_code())



import io
import pytest

from .deps import *

class TestIO(Test):
    def test_reading_stdin(self, internals):
        src = """
und zwar getname
    mein_name ist (bitte? "wie ist dein name? ")
    quasi mein_name
        """
        sin, sout = io.StringIO('alfred\n'), io.StringIO()
        internals.interpreter._ctx.set_stdin(sin)
        internals.interpreter._ctx.set_stdout(sout)
        internals.interpreter._ctx.disable_funny_mode()

        program = internals.compiler.compile(src)
        internals.interpreter.load(program)
        internals.interpreter.call('getname')

        output = sout.getvalue()
        self.assertEqual('wie ist dein name? alfred\n\n', output)



import pytest

from .deps import *

@pytest.fixture
def std():
    obj = Internals()
    src = """use std"""

    program = obj.compiler.compile(src)
    obj.interpreter.load(program)
    return obj

class TestStdlib(Test):
    def test_random(self, std):
        ret = std.interpreter.call('random', [])
        self.assertIsInstance(ret, Number)
        self.assertTrue(0 <= float(ret) <= 1)

    def test_sqrt(self, std):
        ret = std.interpreter.call('sqrt', [16])
        self.assertEqual(4, float(ret))



from ..ast import *
from ..program import *

from .error import CompilerError
from .parser import Parser

class Compiler:
    def __init__(self):
        self._parser = Parser()

    def parser(self):
        return self._parser

    def compile(self, src):
        parsed = self._parser.parse(src)
        return self._finalize(parsed)
    
    def _finalize(self, parsed) -> Program:
        program = Program()
        for item in parsed:
            if isof(item, Declaration):
                if item.is_main():
                    if not program.entry_point() is None:
                        raise CompilerError('main entry point declared twice')
                    program.set_entry_point(item.name())
                program.ident(item.name(), Function(item.args(), item.block()))

            elif isof(item, Use):
                for arg in item.args():
                    program.use(arg)

            else:
                raise CompilerError('unexpected `{}`. only declarations and imports are allowed at top-level.'.format(item))

        return program



from .generic import Keyword

class Marker(Keyword):
    def __str__(self):
        return 'Marker'

class ConstMarker(Marker):
    pass

class MainMarker(Marker):
    pass

class SoMarker(Marker):
    pass



#!/usr/bin/python3

from setuptools import find_packages, setup

setup(
    name='quasicode',
    version='0.0.1',
    author='witling',
    description='the best programming language around.',
    url='https://github.com/witling/quasicode',
    packages=find_packages(),
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: OS Independent',
    ],
    install_requires=[
        'dill'
    ],
    python_requires='>=3.5',
    entry_points={'console_scripts': ['qc=qc.__main__:main']}
)



#!/usr/bin/python3

import inspect
import os
import random
import sys

from os.path import abspath, basename, dirname, expanduser, join, splitext
from shutil import copyfile, rmtree

from qclib import Interpreter, Program, Library
from qclib.library import PyLibrary, init_vlib, get_vlib_modname

def random_id(n=6):
    sample = random.sample('abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ', n)
    return ''.join(sample)

LIBRARY_BUILD = '/tmp/qclibs_{}'.format(random_id())
DEFAULT_IGNORE = ['__pycache__', '__init__.py']

def prepare(build_dir):
    assert not os.path.exists(build_dir)
    os.makedirs(build_dir)

def cleanup(build_dir):
    try:
        rmtree(build_dir)

    except FileNotFoundError:
        print('cannot cleanup dir. not found.')

    except PermissionError:
        print('insufficient permissions to remove file')
        exit()

def build_libraries(folder, ignore=[]):
    import importlib.util

    init_vlib()

    # ignore hidden files
    for script in (path for path in os.listdir(folder) if not path.startswith('.')):
        fname, _ = splitext(script)

        if script in ignore:
            continue

        fpath = join(folder, script)
        modname = get_vlib_modname(fname)

        spec = importlib.util.spec_from_file_location(modname, fpath)
        mod = importlib.util.module_from_spec(spec)

        try:
            spec.loader.exec_module(mod)
            init_vlib(modname, mod)

        except Exception:
            print('Error building', fpath, '. Invalid python code or library missing.')

    subclasses = PyLibrary.__subclasses__()

    for cls in subclasses:
        try:
            if cls.__module__.index(PyLibrary.VIRTUAL_MODULE) != 0:
                continue
        except ValueError:
            continue

        name = cls.__name__.lower()[:-len('library')]
        path = join(LIBRARY_BUILD, '{}{}'.format(name, Program.FEXTC))
        instance = cls()
        if instance.save(path):
            print('Built', name, '...')

def install_libraries(build_dir, dst_dir):
    if not os.path.exists(dst_dir):
        os.makedirs(dst_dir)

    for src in os.listdir(build_dir):
        dst = join(dst_dir, src)
        src = join(build_dir, src)
        ret = copyfile(src, dst)
        print('Installed', splitext(basename(src))[0], 'into', dst, '...')

def process(
    src_dir,
    dst_dir=None,
    ignore=[]
):
    ignore.extend(DEFAULT_IGNORE)
    build_dir = LIBRARY_BUILD
    src_dir = abspath(src_dir)

    if dst_dir is None:
        dst_dir = expanduser(Interpreter.USERLIB_PATH)

    prepare(build_dir)

    build_libraries(src_dir, ignore)
    install_libraries(build_dir, dst_dir)

    cleanup(build_dir)

if __name__ == '__main__':
    main()



import argparse
import sys

try:
    from qclib import *
except:
    from os.path import abspath, dirname, join
    sys.path.insert(0, abspath(join(dirname(__file__), '..')))
    from qclib import *

def argument_parser():
    parser = argparse.ArgumentParser(
        description='Runtime for quasicode. The best language around.'
    )
    subparser = parser.add_subparsers()

    parser_run = subparser.add_parser('run', help='run a quasicode program')
    parser_run.add_argument('program', metavar='PROGRAM', type=str, help='the quasicode program to run.')
    parser_run.add_argument('--listing', action='store_true', default=False, help='printout the program before running.')
    parser_run.add_argument('--debug', action='store_true', default=False, help='enable interpreter debugging.')
    parser_run.add_argument('--nichluschdich', action='store_true', default=False, help='disable the funny mode.')

    parser_install = subparser.add_parser('install', help='install a library.')
    parser_install.add_argument('library', metavar='LIBRARY', type=str, help='path to the library to install.')
    parser_install.add_argument('--ignore', type=str, default=[], nargs='+', help='files to ignore')

    return parser

def run(args):
    program = Program.load(args.program)
    interpreter = Interpreter()
    
    if args.listing:
        print('=== Listing', '=' * 8)
        print(program)
        print('=' * 20)

    if args.debug:
        from pudb import set_trace
        set_trace()

    if args.nichluschdich:
        interpreter.disable_funny_mode()

    interpreter.load(program)
    interpreter.run()

def install(args):
    from .makelibs import process
    from os.path import abspath

    process(
        src_dir=args.library,
        ignore=args.ignore
    )

def main():
    args = argument_parser().parse_args()

    if 'program' in args:
        run(args)

    elif 'library' in args:
        install(args)

    else:
        print('no command specified.')

if __name__ == '__main__':
    main()



from .generic import *

class Branch(NestedStatement, Runnable):
    def __init__(self):
        self._branches = []
        self._default_branch = None

    def branches(self):
        return [*self._branches, (None, self._default_branch)] if self._default_branch else self._branches

    def add_branch(self, condition, block):
        self._branches.append((condition, block))

    def set_default_branch(self, block):
        self._default_branch = block

    def run(self, ctx):
        for condition, block in self._branches:
            if condition.run(ctx):
                block.run(ctx)
                break
        else:
            if self._default_branch:
                self._default_branch.run(ctx)

    def __str__(self):
        fixed = list(map(lambda x: '| {} -> {}'.format(x[0], x[1]), self._branches))
        if self._default_branch:
            fixed.append('| -> {}'.format(self._default_branch))
        return '\n'.join(fixed)

class If(Branch):
    def __init__(self):
        super().__init__()

    def __str__(self):
        return 'if:\n' + super().__str__()

class Elif(Branch):
    def __init__(self):
        super().__init__()

    def __str__(self):
        return 'elif'

class Else(Branch):
    def __init__(self):
        super().__init__()

    def __str__(self):
        return 'else'



from os.path import abspath, dirname, join
import pytest
import sys

sys.path.insert(0, abspath(join(dirname(__file__), '..')))

from qclib import *

class Test:
    def assertIsInstance(self, obj, cls):
        self.assertTrue(isinstance(obj, cls))

    def assertEqual(self, expected, got):
        self.assertTrue(expected == got)

    def assertFalse(self, expr):
        self.assertTrue(not expr)

    def assertTrue(self, expr):
        assert expr

class Internals:
    def __init__(self):
        self.interpreter = Interpreter()
        self.compiler = Compiler()
        self.parser = self.compiler.parser()

@pytest.fixture
def internals():
    return Internals()



from .generic import *
from .value import Ident, Liste

class Construct(Parameterized):
    def __init__(self, ty, init):
        self._ty = ty
        self._init = init

    def run(self, ctx):
        if isof(self._ty, Liste):
            liste = self._ty
            for item in self._init:
                if isof(item, Ident):
                    liste._val.append(item.run(ctx))
                else:
                    liste._val.append(item)
            return liste
        raise Exception('unexpected type')

class Readin(Statement, Parameterized):
    RESPONSES = ['verstehe.', 'gut.', 'okay.', 'perfekt.']

    def __init__(self):
        Statement.__init__(self)
        Parameterized.__init__(self)

    def run(self, ctx):
        question = ' '.join(map(str, self.args()))

        ctx.stdout().write(question)
        ctx.stdout().flush()

        ret = ctx.stdin().readline()

        if ctx.is_funny_mode():
            from random import choice
            ctx.stdout().write('{}\n'.format(choice(Readin.RESPONSES)))
            ctx.stdout().flush()

        return ret

    def __str__(self):
        return 'stdin_read {}'.format(' '.join(map(str, self.args())))

class Print(Statement, Parameterized):
    def __init__(self):
        Statement.__init__(self)
        Parameterized.__init__(self)

    def run(self, ctx):
        msg = ' '.join(map(lambda x: str(x.run(ctx)), self._args)) + '\n'
        ctx.stdout().write(msg)
        ctx.stdout().flush()

    def __str__(self):
        return 'print {}'.format(' '.join(map(str, self._args)))

class Debug(Statement):
    def __init__(self):
        super().__init__()

    def run(self, ctx):
        from pudb import set_trace
        set_trace()

    def __str__(self):
        return 'debug'

class Use(Statement, Parameterized):
    def __init__(self):
        Statement.__init__(self)
        Parameterized.__init__(self)

    def run(self, ctx):
        modname = self.args()[0]
        raise Exception('local use is not implemented :(')

    def __str__(self):
        return 'use'

class Exit(Statement):
    def __init__(self):
        super().__init__()

    def __str__(self):
        return 'exit'

class Nop(NestedStatement):
    def __init__(self):
        super().__init__()

    def run(self, ctx):
        ctx.fun()

    def __str__(self):
        return 'nop'

class Raise(NestedStatement):
    def __init__(self):
        super().__init__()

    def __str__(self):
        return 'raise'

class Return(Statement, Parameterized):
    def __init__(self):
        Statement.__init__(self)
        Parameterized.__init__(self)

    def run(self, ctx):
        ctx.set_return(self.args()[0].run(ctx))

    def __str__(self):
        return 'return {}'.format(' '.join(map(str, self._args)))

class Repeat(NestedStatement):
    def __init__(self):
        super().__init__()
        self._broken = False

    def end(self):
        self._broken = True

    def run(self, ctx):
        ctx.push_loop(self)
        while not self._broken:
            for step in self.block():
                step.run(ctx)
                if self._broken:
                    break
        ctx.pop_loop()

    def __str__(self):
        return super().__str__()

class Break(NestedStatement):
    def __init__(self):
        super().__init__()

    def run(self, ctx):
        ctx.last_loop().end()

    def __str__(self):
        return 'break'
